const sortBy = require('sort-by');

const REVERSED = 1;

/*
 * Filter
 */

const createFilterPredicate = expression => (
  entry => Object.keys(expression).every(key => entry[key] === expression[key])
);

const getFilterPredicate = expression => (
  typeof expression === 'function' ? expression : createFilterPredicate(expression)
);

/*
 * Exclude
 */

const createExcludePredicate = expression => (
  entry => Object.keys(expression).every(key => entry[key] !== expression[key])
);

const getExcludePredicate = expression => (
  typeof expression === 'function' ? expression : createExcludePredicate(expression)
);

/*
 * Mixin
 */

class SortedEntriesGenerator {

  constructor(queryset, sortComparator) {
    this.sortComparator = sortComparator;
    this.parent = queryset.parent;
  }

  * entries() {
    for (const [, entry] of this.parent.toArray().sort(this.sortComparator).entries()) {
      yield entry;
    }
  }

}

const withNonEntriesGenerator = Base => class extends Base {

  constructor(...args) {
    super(...args);

    this.source = [];
  }

};

const withQueries = Base => class extends Base {

  constructor({QuerySetConstructor, parent = null, predicate = null, sortComparator = null}) {
    super({parent, QuerySetConstructor});
    this.predicate = predicate;
    this.source = sortComparator ? new SortedEntriesGenerator(this, sortComparator) : null;
  }

  validate(entry) {
    return !this.predicate || this.predicate(entry);
  }

  getFallbackSource(source = null) {
    return source || this.parent;
  }

  * entries() {
    const source = this.getFallbackSource(this.source);

    for (const entry of source.entries()) {
      if (this.validate(entry)) yield entry;
    }
  }

  filter(expression) {
    return this.clone({predicate: getFilterPredicate(expression)});
  }

  exclude(expression) {
    return this.clone({predicate: getExcludePredicate(expression)});
  }

  sort(sortComparator) {
    return this.clone({sortComparator});
  }

  orderBy(...fields) {
    return this.sort(sortBy(...fields));
  }

  reverse() {
    return this.clone({sortComparator: () => REVERSED});
  }

  none() {
    return this.clone(null, withNonEntriesGenerator(this.QuerySetConstructor));
  }

  all() {
    return this.clone();
  }

};

module.exports = {
  withQueries,
  getFilterPredicate,
  getExcludePredicate,
};
